/*
 * Copyright (c) 2022 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include "common.h"

#include <zephyr/kernel.h>
#include <zephyr/device.h>

#include <zephyr/ipc/ipc_service.h>
#include <hal/nrf_reset.h>
#include <string.h>

#include <stdio.h>
#include <stdlib.h>

#include <zephyr/drivers/gpio.h>
#include <zephyr/drivers/uart.h>
#include <zephyr/usb/usb_device.h>
#include <zephyr/usb/usbd.h>

#define NRFX_LOG_MODULE EXAMPLE
#define NRFX_EXAMPLE_CONFIG_LOG_ENABLED 1
#define NRFX_EXAMPLE_CONFIG_LOG_LEVEL 3
#define MSG_SIZE 32
#include <nrfx_log.h>
#include <nrfx_gpiote.h>

#define SPIKE_MODE 0
#define RAW_MODE 1

#define SPIKE_LENGTH 210
#define RAW_LENGTH 250

unsigned char _mode = SPIKE_MODE;
unsigned char _channel[4];

const struct device *ipc0_instance;
struct ipc_ept ep;

K_SEM_DEFINE(bound_sem, 0, 1);
K_MSGQ_DEFINE(uart_msgq, MSG_SIZE, 10, 4);
static char rx_buf[MSG_SIZE];
static int rx_buf_pos;

#include <zephyr/sys/ring_buffer.h>

#define USB_DATA_LENGTH 15750

#define ESB_RING_BUFFER_SIZE (USB_DATA_LENGTH)

RING_BUF_DECLARE(esb_ring_buf, ESB_RING_BUFFER_SIZE);

uint8_t USB_Send_Buffer[USB_DATA_LENGTH];

K_EVENT_DEFINE(usb_event);
K_EVENT_DEFINE(read_event);

const struct device *usb_dev;

uint8_t usb_get_buffer[50] = {0};
size_t usb_recv_len;

static void interrupt_handler(const struct device *dev, void *user_data)
{
    ARG_UNUSED(user_data);

    while (uart_irq_update(dev) && uart_irq_is_pending(dev))
    {

        if (uart_irq_rx_ready(dev))
        {

            usb_recv_len = uart_fifo_read(dev, usb_get_buffer, sizeof(usb_get_buffer));
            if (usb_get_buffer[0] == 0x03)
            {
                k_event_set(&read_event, 0x01);
                return;
            }

            if (usb_get_buffer[0] == 0x00 || usb_get_buffer[0] == 0x01 || usb_get_buffer[0] == 0x02)
            {
                _mode = usb_get_buffer[0];
                _channel[0] = usb_get_buffer[1];
                _channel[1] = usb_get_buffer[2];
                _channel[2] = usb_get_buffer[3];
                _channel[3] = usb_get_buffer[4];
                k_event_set(&usb_event, 0x01);
                return;
            }

            // k_event_set(&usb_event, 0x01);

            // NRFX_LOG_INFO("Get - %s.", usb_get_buffer);

            // NRFX_EXAMPLE_LOG_PROCESS();

            // if (recv_len < 0) {
            // 	NRFX_LOG_INFO("Failed to read UART FIFO");
            //     NRFX_EXAMPLE_LOG_PROCESS();
            // 	recv_len = 0;
            // };
            // uart_fifo_fill(dev, Send_Buffer, SEND_BUFFER_SIZE);
            // uart_irq_tx_disable(dev);
        }
    }
}

static void ep_bound(void *priv)
{
    k_sem_give(&bound_sem);
    LOG_INF("Ep bounded");
}

static void ep_recv(const void *data, size_t len, void *priv)
{
    if (len > 0)
    {
        if(len == (_mode == SPIKE_MODE ? SPIKE_LENGTH : RAW_LENGTH) && ring_buf_space_get(&esb_ring_buf) >= len) 
        {
            ring_buf_put(&esb_ring_buf, data, len);
        }
        else ring_buf_reset(&esb_ring_buf);
    }
    
}

static struct ipc_ept_cfg ep_cfg = {
    .cb = {
        .bound = ep_bound,
        .received = ep_recv,
    },
};

void ipc_init()
{
    int ret;

    LOG_INF("IPC-service HOST demo started");

    ipc0_instance = DEVICE_DT_GET(DT_NODELABEL(ipc0));

    ret = ipc_service_open_instance(ipc0_instance);
    if ((ret < 0) && (ret != -EALREADY))
    {
        LOG_ERR("ipc_service_open_instance() failure");
        return ret;
    }

    ret = ipc_service_register_endpoint(ipc0_instance, &ep, &ep_cfg);
    if (ret < 0)
    {
        LOG_ERR("ipc_service_register_endpoint() failure");
        return ret;
    }
}

void usb_init()
{
    uint32_t baudrate, dtr = 0U;
    int ret;

    usb_dev = DEVICE_DT_GET_ONE(zephyr_cdc_acm_uart);
    if (!device_is_ready(usb_dev))
    {
        NRFX_LOG_ERROR("CDC ACM device not ready");
        return 0;
    }
    ret = usb_enable(NULL);
    if (ret != 0)
    {
        NRFX_LOG_ERROR("Failed to enable USB");
        return 0;
    }
    
    uart_irq_callback_set(usb_dev, interrupt_handler);

    uart_irq_rx_enable(usb_dev);
}

int main(void)
{
    NRFX_EXAMPLE_LOG_INIT();

    ipc_init();

    usb_init();

    uint8_t _send[] = {0x00, 0x00, 0x00, 0x00, 0x00, 0xff};
    uint32_t events;
    NRFX_LOG_INFO("USB_Get_Thread Start.");
    NRFX_EXAMPLE_LOG_PROCESS();

    while (1)
    {
        events = k_event_wait(&read_event, 0x01, true, K_MSEC(500));
        if (events != 0)
        {
            size_t _data_length = (_mode == SPIKE_MODE ? SPIKE_LENGTH : RAW_LENGTH);

            for (size_t i = 0; i < (USB_DATA_LENGTH / _data_length); i++)
            {
            re_read:
                if (ring_buf_is_empty(&esb_ring_buf))
                {
                    k_sleep(K_MSEC(5));
                    goto re_read;
                }
                ring_buf_get(&esb_ring_buf, &(USB_Send_Buffer[i * _data_length]), _data_length);
          
            }
            NRFX_EXAMPLE_LOG_PROCESS();
            while (usb_transfer_is_busy(((struct usb_cfg_data *)usb_dev->config)->endpoint[2].ep_addr))
                ;
            usb_transfer_sync(((struct usb_cfg_data *)usb_dev->config)->endpoint[2].ep_addr, USB_Send_Buffer, USB_DATA_LENGTH, USB_TRANS_WRITE);
        }
    }

    return 0;
}

static int network_gpio_allow(void)
{
    // NRF_P0_S->PIN_CNF[4] = (GPIO_PIN_CNF_MCUSEL_NetworkMCU << GPIO_PIN_CNF_MCUSEL_Pos);
    // NRF_P0_S->PIN_CNF[5] = (GPIO_PIN_CNF_MCUSEL_NetworkMCU << GPIO_PIN_CNF_MCUSEL_Pos);
    // NRF_P0_S->PIN_CNF[28] = (GPIO_PIN_CNF_MCUSEL_NetworkMCU << GPIO_PIN_CNF_MCUSEL_Pos);
    // NRF_P0_S->PIN_CNF[26] = (GPIO_PIN_CNF_MCUSEL_NetworkMCU << GPIO_PIN_CNF_MCUSEL_Pos);

    NRF_P0_S->PIN_CNF[25] = (GPIO_PIN_CNF_MCUSEL_NetworkMCU << GPIO_PIN_CNF_MCUSEL_Pos);
    NRF_P0_S->PIN_CNF[29] = (GPIO_PIN_CNF_MCUSEL_NetworkMCU << GPIO_PIN_CNF_MCUSEL_Pos);
    NRF_P0_S->PIN_CNF[30] = (GPIO_PIN_CNF_MCUSEL_NetworkMCU << GPIO_PIN_CNF_MCUSEL_Pos);

    return 0;
}

SYS_INIT(network_gpio_allow, PRE_KERNEL_1, CONFIG_KERNEL_INIT_PRIORITY_OBJECTS);

void USB_Push_Thread()
{
    uint8_t _send[] = {0x00, 0x00, 0x00, 0x00, 0x00, 0xff};
    uint32_t events;
    NRFX_LOG_INFO("USB_Push_Thread Start.");
    NRFX_EXAMPLE_LOG_PROCESS();

    while (1)
    {
        events = k_event_wait(&usb_event, 0x01, true, K_MSEC(500));
        if (events != 0)
        {
            _send[0] = _mode ? 0x01 : 0x00;
            _send[1] = _channel[0];
            _send[2] = _channel[1];
            _send[3] = _channel[2];
            _send[4] = _channel[3];
            ipc_service_send(&ep, _send, 6);
            ring_buf_reset(&esb_ring_buf);
            NRFX_LOG_INFO("Send %x %x %x %x %x", _send[0], _send[1], _send[2], _send[3], _send[4]);
            NRFX_EXAMPLE_LOG_PROCESS();
        }
    }
}
K_THREAD_DEFINE(usb_push_id, 1024, USB_Push_Thread, NULL, NULL, NULL, 7, 0, 0);

// void USB_Get_Thread()
// {

// }

// K_THREAD_DEFINE(usb_get_id, 1024, USB_Get_Thread, NULL, NULL, NULL, 7, 0, 5);