/*
 * Copyright (c) 2022 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/kernel.h>
#include <zephyr/device.h>

#include <zephyr/ipc/ipc_service.h>

#include "common.h"

#include <zephyr/logging/log.h>

#include <zephyr/drivers/clock_control.h>
#include <zephyr/drivers/clock_control/nrf_clock_control.h>
#include <esb.h>
#include <nrfx_spim.h>

#include <zephyr/drivers/gpio.h>

#define SPIKE_MODE 0
#define RAW_MODE 1

#define SPIKE_LENGTH 210
#define RAW_LENGTH 250

#if !DT_NODE_EXISTS(DT_NODELABEL(fpga_gpios))
#error "Overlay for fpga_gpios node not properly defined."
#endif

// #define _DELAY_()                    \
//     do                               \
//     {                                \
//         uint32_t __i__ = 0x00FFFFFF; \
//         while (__i__--)              \
//             ;                        \
//     } while (0ULL)

static const struct gpio_dt_spec full_gpio = GPIO_DT_SPEC_GET_BY_IDX_OR(DT_NODELABEL(fpga_gpios), gpios, 0, {0});
static const struct gpio_dt_spec clear_gpio = GPIO_DT_SPEC_GET_BY_IDX_OR(DT_NODELABEL(fpga_gpios), gpios, 1, {0});
static const struct gpio_dt_spec mode_gpio = GPIO_DT_SPEC_GET_BY_IDX_OR(DT_NODELABEL(fpga_gpios), gpios, 2, {0});

#define UPDATE_PERIOD 30 // One data transfer after every nine data transfers carries an ACK message

unsigned char _mode = SPIKE_MODE;

static struct esb_payload rx_payload;
static struct esb_payload tx_payload = {.pipe = 0, .length = 5, .noack = false, .data = {0, 0, 0, 0, 0}};
struct esb_config config = ESB_DEFAULT_CONFIG;

struct ipc_ept ep;

LOG_MODULE_REGISTER(remote, LOG_LEVEL_INF);

K_SEM_DEFINE(bound_sem, 0, 1);
K_SEM_DEFINE(DATA_PARSING_SEM, 0, 1);

uint32_t _count_ = UPDATE_PERIOD;

// #define MOSI_PIN_MASTER 4
// #define MISO_PIN_MASTER 5
// #define SCK_PIN_MASTER 28
// #define SS_PIN_MASTER 26

// #define SPIM_INST_IDX 0
// nrfx_spim_t spim_inst = NRFX_SPIM_INSTANCE(SPIM_INST_IDX);
// unsigned char m_tx_buffer_master[2];
// nrfx_spim_xfer_desc_t spim_xfer_desc_tx = NRFX_SPIM_XFER_TX(m_tx_buffer_master, 2);

void event_handler(struct esb_evt const *event)
{
    switch (event->evt_id)
    {
    case ESB_EVENT_TX_SUCCESS:
        break;
    case ESB_EVENT_TX_FAILED:
        break;
    case ESB_EVENT_RX_RECEIVED:
        esb_read_rx_payload(&rx_payload);
        gpio_pin_toggle_dt(&mode_gpio);
        // k_sem_give(&DATA_PARSING_SEM);

        // _count_--;
        // if (_count_ == 0)
        // {
        //     esb_write_payload(&tx_payload);
        //     _count_ = UPDATE_PERIOD;
        // }

        break;
    }
}

static void ep_bound(void *priv)
{
    k_sem_give(&bound_sem);
    LOG_INF("Ep bounded");
}

static void ep_recv(const void *data, size_t len, void *priv)
{
    if(*(unsigned char *)data == 0x00 || *(unsigned char *)data == 0x01)
        _mode = *(unsigned char *)data;
    memcpy(tx_payload.data, data, len);
    // gpio_pin_set_dt(&clear_gpio, !(1));
    // _DELAY_();
    // gpio_pin_set_dt(&clear_gpio, !(0));
    // gpio_pin_set_dt(&mode_gpio, !(_mode));
    // k_sem_reset(&DATA_PARSING_SEM);
}

static struct ipc_ept_cfg ep_cfg = {
    .cb = {
        .bound = ep_bound,
        .received = ep_recv,
    },
};

int clocks_start(void)
{
    int err;
    int res;
    struct onoff_manager *clk_mgr;
    struct onoff_client clk_cli;

    clk_mgr = z_nrf_clock_control_get_onoff(CLOCK_CONTROL_NRF_SUBSYS_HF);
    if (!clk_mgr)
    {
        LOG_ERR("Unable to get the Clock manager");
        return -ENXIO;
    }

    sys_notify_init_spinwait(&clk_cli.notify);

    err = onoff_request(clk_mgr, &clk_cli);
    if (err < 0)
    {
        LOG_ERR("Clock request failed: %d", err);
        return err;
    }

    do
    {
        err = sys_notify_fetch_result(&clk_cli.notify, &res);
        if (!err && res)
        {
            LOG_ERR("Clock could not be started: %d", res);
            return res;
        }
    } while (err);

    LOG_DBG("HF clock started");
    return 0;
}

int esb_initialize(void)
{
    int err;

    uint8_t base_addr_0[4] = {0xA0, 0xA0, 0xA0, 0xA0};
    uint8_t base_addr_1[4] = {0x00, 0x00, 0x00, 0x00};
    uint8_t addr_prefix[]  = {0xA0};

    config.protocol = ESB_PROTOCOL_ESB_DPL;
    config.retransmit_delay = 150;
    config.bitrate = ESB_BITRATE_2MBPS;
    config.mode = ESB_MODE_PRX;
    config.event_handler = event_handler;
    config.payload_length = 252;
    config.tx_mode = ESB_TXMODE_AUTO;
    config.use_fast_ramp_up = true;
    config.selective_auto_ack = true;
    config.tx_output_power = ESB_TX_POWER_0DBM;
    config.retransmit_count = 0;
    config.crc = ESB_CRC_OFF;

    err = esb_init(&config);
    if (err)
    {
        return err;
    }

    err = esb_set_base_address_0(base_addr_0);
    if (err)
    {
        return err;
    }

    err = esb_set_base_address_1(base_addr_1);
    if (err)
    {
        return err;
    }

    err = esb_set_prefixes(addr_prefix, ARRAY_SIZE(addr_prefix));
    if (err)
    {
        return err;
    }

    return 0;
}

void init()
{
    const struct device *ipc0_instance;
    int ret;

    LOG_INF("IPC-service REMOTE demo started");

    ipc0_instance = DEVICE_DT_GET(DT_NODELABEL(ipc0));

    ret = clocks_start();
    if (ret)
    {
        return;
    }

    ret = ipc_service_open_instance(ipc0_instance);
    if ((ret < 0) && (ret != -EALREADY))
    {
        LOG_ERR("ipc_service_open_instance() failure");
        return;
    }

    ret = ipc_service_register_endpoint(ipc0_instance, &ep, &ep_cfg);
    if (ret != 0)
    {
        LOG_ERR("ipc_service_register_endpoint() failure");
        return;
    }

    ret = esb_initialize();
    if (ret)
    {
        LOG_ERR("ESB initialization failed, err %d", ret);
        return;
    }

    ret = esb_start_rx();
    if (ret)
    {
        LOG_ERR("RX setup failed, err %d", ret);
        return;
    }

    LOG_INF("Initialization complete");
    LOG_INF("Mode change In %s", config.mode == ESB_MODE_PRX ? "PRX" : "PTX");

    k_sem_take(&bound_sem, K_FOREVER);
}

// void spim_init()
// {
//     nrfx_err_t status;
//
//     nrfx_spim_config_t spim_config = NRFX_SPIM_DEFAULT_CONFIG(SCK_PIN_MASTER,
//                                                               MOSI_PIN_MASTER,
//                                                               MISO_PIN_MASTER,
//                                                               SS_PIN_MASTER);
//
//     spim_config.mode = NRF_SPIM_MODE_1;
//
//     void *p_context = "Some context";
//     status = nrfx_spim_init(&spim_inst, &spim_config, NULL, p_context);
//     NRFX_ASSERT(status == NRFX_SUCCESS);
//
// #if defined(__ZEPHYR__)
//     IRQ_DIRECT_CONNECT(NRFX_IRQ_NUMBER_GET(NRF_SPIM_INST_GET(SPIM_INST_IDX)), IRQ_PRIO_LOWEST,
//                        NRFX_SPIM_INST_HANDLER_GET(SPIM_INST_IDX), 0);
// #endif
//
//     nrf_spim_enable(spim_inst.p_reg);
// }

// void spim_send()
// {
//     for(unsigned int i = 0; i < (rx_payload.length / 2); i++)
//     {
//         nrf_gpio_pin_write(SS_PIN_MASTER, 0);
//         nrf_spim_tx_buffer_set(spim_inst.p_reg, &(rx_payload.data[2 * i]), 2);
//         nrf_spim_event_clear(spim_inst.p_reg, NRF_SPIM_EVENT_ENDTX);
//         nrf_spim_task_trigger(spim_inst.p_reg, NRF_SPIM_TASK_START);
//         while(!nrf_spim_event_check(spim_inst.p_reg, NRF_SPIM_EVENT_ENDTX));
//         nrf_gpio_pin_write(SS_PIN_MASTER, 1);
//     }
// }

void gpio_init()
{
    gpio_is_ready_dt(&clear_gpio);
    gpio_pin_configure_dt(&clear_gpio, GPIO_OUTPUT_ACTIVE);
    gpio_pin_set_dt(&clear_gpio, !(0));

    gpio_is_ready_dt(&mode_gpio);
    gpio_pin_configure_dt(&mode_gpio, GPIO_OUTPUT_ACTIVE);
    gpio_pin_set_dt(&mode_gpio, !(0));

    gpio_is_ready_dt(&full_gpio);
    gpio_pin_configure_dt(&full_gpio, GPIO_INPUT);
}

int main(void)
{
    init();

    gpio_init();

    // spim_init();

    // esb_write_payload(&tx_payload);

    //while (1)
    //{
        // gpio_pin_toggle_dt(&mode_gpio);
        // if (k_sem_take(&DATA_PARSING_SEM, K_FOREVER) == 0)
        // {
        //     if (_mode == SPIKE_MODE && rx_payload.length != SPIKE_LENGTH)
        //         continue;
        //     else if (_mode == RAW_MODE && rx_payload.length != RAW_LENGTH)
        //         continue;

            // ipc_service_send(&ep, rx_payload.data, rx_payload.length);
            // gpio_pin_toggle_dt(&mode_gpio);
            // if(gpio_pin_get_dt(&full_gpio))
            // spim_send();
        // }
    //}

    // #define BUFF_SIZES 250

    //     unsigned char bufferss[BUFF_SIZES] = {[0] = 0x00, [1] = 0x00, [2] = 0x33, [3] = 0x33, [4] = 0x00, [5] = 0x00, [6] = 0x33, [7] = 0x33, [8] = 0x00, [9] = 0x00,};

    //     while(1)
    //     {

    //         for(unsigned int i = 0; i < (BUFF_SIZES / 2); i++)
    //         {
    //             nrf_gpio_pin_write(SS_PIN_MASTER, 0);
    //             nrf_spim_tx_buffer_set(spim_inst.p_reg, &(bufferss[2 * i]), 2);
    //             nrf_spim_event_clear(spim_inst.p_reg, NRF_SPIM_EVENT_ENDTX);
    //             nrf_spim_task_trigger(spim_inst.p_reg, NRF_SPIM_TASK_START);
    //             while(!nrf_spim_event_check(spim_inst.p_reg, NRF_SPIM_EVENT_ENDTX));
    //             nrf_gpio_pin_write(SS_PIN_MASTER, 1);
    //         }

    //         k_sleep(K_MSEC(500));
    //     }

    return 0;
}

// void Data_Parsing_Thread(void *p1, void *p2, void *p3)
// {

//     while (1)
//     {
//         gpio_pin_toggle_dt(&mode_gpio);
//         if (k_sem_take(&DATA_PARSING_SEM, K_FOREVER) == 0)
//         {
//             if (_mode == SPIKE_MODE && rx_payload.length != SPIKE_LENGTH)
//                 continue;
//             else if (_mode == RAW_MODE && rx_payload.length != RAW_LENGTH)
//                 continue;

//             ipc_service_send(&ep, rx_payload.data, rx_payload.length);
// gpio_pin_toggle_dt(&mode_gpio);
// if(gpio_pin_get_dt(&full_gpio))
// spim_send();
//         }
//     }
// }

// K_THREAD_DEFINE(DATA_PARSING_THREAD_ID, 512, Data_Parsing_Thread, NULL, NULL, NULL, 2, 0, 0);